(*
 * SNU 4190.310 Programming Languages (Fall 2008)
 *
 * K- Interpreter I
 *)

(*
 * Location
 *)
signature LOC =
  sig
    (*
     * Base Address
     *)
    structure Base :
      sig
        (* type of bass address *)
        type t

        (* NULL *)
        val null : t

        (* successor *)
        val succ : t -> t

        (* equal b1 b2 tests equality of b1 and b2 *)
        val equal : t -> t -> bool
      end

    (* type of location *)
    type t = Base.t * tyoffset
    and tyoffset = int

    exception Not_allowed

    (* loc_of_base b returns the first (offset 0) location of b *)
    val loc_of_base : Base.t -> t

    (* base l returns the base address of l. *)
    val base : t -> Base.t

    (* offset l return the offset of l. *)
    val offset : t -> int

    (* addi l i adds i to the offset of l. *)
    val addi : t -> int -> t

    (* subi l i subtracts i from the offset of l. *)
    val subi : t -> int -> t

    (* sub x y returns (offset x) - (offset y), or raises Not_allowed if they
       have different base addresses. *)
    val sub : t -> t -> tyoffset

    (* equal l r tests equality of l and r, or raises Not_allowed if they have
       different base addresses. *)
    val equal : t -> t -> bool

    (* less l r returns (offset l) < (offset r), or raises Not_allowed if they
       have different base addresses. *)
    val less : t -> t -> bool
  end

structure Loc : LOC =
  struct
    structure Base =
      struct
        type t = int

        val null = 0

        fun succ b = b + 1

        fun equal l r = (l = r)
      end

    type t = Base.t * tyoffset
    and tyoffset = int

    exception Not_allowed

    fun loc_of_base base = (base, 0)

    fun base (base, _) = base

    fun offset (_, offset) = offset

    fun addi (base, offset) n = (base, offset + n)

    fun subi (base, offset) n = (base, offset - n)

    fun sub (base, offset) (base', offset') =
        if Base.equal base base' then offset - offset' else raise Not_allowed

    fun equal (base, offset) (base', offset') =
        (Base.equal base base') && (offset = offset')

    fun less (base, offset) (base', offset') =
        if Base.equal base base' then offset < offset' else raise Not_allowed
  end

(*
 * Memory
 *)
signature MEM =
  sig
    (* type of memory from Loc.t to 'a *)
    type 'a t

    exception Not_allocated

    (* empty memory *)
    val empty : 'a t

    (* allocate m s returns an Loc.t to a fresh (uninitialized) buffer of size
       s and a memory containing the same bindings as m, plus a binding of
       the Loc.t to the first location of allocated buffer. *)
    val allocate : 'a t -> int -> Loc.t * 'a t

	(* "new m" equals to "allocate m 1" *)
	val new : 'a t -> Loc.t * 'a t

    (* store m l v returns a memory containing the same bindings as m, with l
       bound to v, or raises Not_allocated if m doesn't contain a binding for
       l. *)
    val store : 'a t -> Loc.t -> 'a -> 'a t

    (* load m l returns the current binding of l in m, or raises Not_found if
       no such binding exists. *)
    val load : 'a t -> Loc.t -> 'a

    (* is_allocated m l returns true if m contains a binding for l, and false
       otherwise. *)
    val is_allocated : 'a t -> Loc.t -> bool
  end

structure Mem : MEM =
  struct
    type 'a t = Loc.Base.t * (Loc.t, 'a option) Map.t

    exception Not_allocated
	exception Not_initialized

    val empty = (Loc.Base.null, Map.empty)

    fun allocate (base, mem) s =
        let
          fun range f t = if f < t then f :: (range (f + 1) t) else []

          val base' = Loc.Base.succ base
          val l = Loc.loc_of_base base'
        in
          (l, (base', List.fold_left
                        (fn m o => Map.add (Loc.addi l o) None m)
                        mem
                        (range 0 s)))
        end

	fun new m = allocate m 1

    fun is_allocated (_, mem) l = Map.mem l mem

    fun store (m as (base, mem)) l v =
        if is_allocated m l then (base, Map.add l (Some v) mem)
        else raise Not_allocated

    fun load (_, mem) l =
        let val content = (Map.find l mem) in 
				case content of
				None => raise Not_initialized
				| Some v => v end
		handle Not_found => raise Not_allocated
  end

(*
 * Environment
 *)
signature ENV =
  sig
    (* type of environment from 'a to 'b *)
    type ('a, 'b) t

    exception Not_bound

    (* empty environment *)
    val empty : ('a, 'b) t

    (* bind e x l returns an environment containing the same bindings as e,
       plus a binding of x to l. If x was already bound in e, its previous
       binding disappears. *)
    val bind : ('a, 'b) t -> 'a -> 'b -> ('a, 'b) t

    (* lookup e x returns the current binding of x in e, or raises Not_bound if
       no such binding exists. *)
    val lookup : ('a, 'b) t -> 'a -> 'b

    (* is_bound e x returns true if e contains a binding for x, and false
       otherwise. *)
    val is_bound : ('a, 'b) t -> 'a -> bool
  end

structure Env : ENV =
  struct
    type ('a, 'b) t = ('a, 'b) Map.t

    exception Not_bound

    val empty = Map.empty

    fun bind e x a = Map.add x a e

    fun lookup e x = (Map.find x e) handle Not_found => raise Not_bound

    fun is_bound e x = Map.mem x e
  end

(*
 * K- Interpreter
 *)
signature KMINUS =
  sig
    exception Error of string
    type id = string
    type exp =
        NUM of int | TRUE | FALSE | UNIT
      | VAR of id
      | ADD of exp * exp
      | SUB of exp * exp
      | MUL of exp * exp
      | DIV of exp * exp
      | EQUAL of exp * exp
      | LESS of exp * exp
      | NOT of exp
      | ASSIGN of id * exp            (* assgin to variable *)
      | SEQ of exp * exp              (* sequence *)
      | IF of exp * exp * exp         (* if-then-else *)
      | WHILE of exp * exp            (* while loop *)
      | FOR of id * exp * exp * exp   (* for loop *)
      | LETV of id * exp * exp        (* variable binding *)
      | LETF of id * id * exp * exp   (* procedure binding *)
      | CALLV of id * exp             (* call by value *)
      | CALLR of id * id              (* call by referenece *)
      | READ of id
      | WRITE of exp
      | RECORD of (id * exp) list     (* record *)
      | FIELD of exp * id             (* record field access *)
      | ASSIGNFIELD of exp * id * exp (* assign to recoed field *)
      | MALLOC of exp                 (* malloc *)
      | AMPER of id                   (* address of id *)
      | AMPERFIELD of exp * id        (* address of record field *)
      | STAR of exp                   (* dereference *)
      | ASSIGNSTAR of exp * exp       (* assign to locatio *)
    type program = exp
    type memory
    type env
    type value
    val emptyMemory : memory
    val emptyEnv : env
    val run : memory * env * program -> value
  end

structure K : KMINUS =
  struct
    exception Error of string

    type id = string

    type exp =
        NUM of int | TRUE | FALSE | UNIT
      | VAR of id
      | ADD of exp * exp
      | SUB of exp * exp
      | MUL of exp * exp
      | DIV of exp * exp
      | EQUAL of exp * exp
      | LESS of exp * exp
      | NOT of exp
      | ASSIGN of id * exp
      | SEQ of exp * exp
      | IF of exp * exp * exp
      | WHILE of exp * exp
      | FOR of id * exp * exp * exp
      | LETV of id * exp * exp
      | LETF of id * id * exp * exp
      | CALLV of id * exp
      | CALLR of id * id
      | READ of id
      | WRITE of exp
      | RECORD of (id * exp) list 
      | FIELD of exp * id
      | ASSIGNFIELD of exp * id * exp
      | MALLOC of exp
      | AMPER of id 
      | AMPERFIELD of exp * id
      | STAR of exp
      | ASSIGNSTAR of exp * exp
    type program = exp

    type value =
        Num of int
      | Bool of bool
      | Location of Loc.t
	  | Record (* XXX *)
      | Unit
      | Bot

    fun value_int Num n = n
      | value_int Bool _ = raise (Error "Bool type is used as Num type")
      | value_int Location _ = raise (Error "Location type is used as Num type")
      | value_int Unit = raise (Error "Unit type is used as Num type")
      | value_int Bot = raise (Error "not initialized")

    fun value_bool Bool b = b
      | value_bool Num _ = raise (Error "Num type is used as Bool type")
      | value_bool Location _ = raise (Error "Location type is used as Bool type")
      | value_bool Unit = raise (Error "Unit type is used as Bool type")
      | value_bool Bot = raise (Error "not initialized")

    fun value_loc Location l = l
      | value_loc Num _ = raise (Error "Num type is used as Location type")
      | value_loc Bool _ = raise (Error "Bool type is used as Location type")
      | value_loc Unit = raise (Error "Unit type is used as Location type")
      | value_loc Bot = raise (Error "not initialized")

    fun value_unit Unit = ()
      | value_unit Num _ = raise (Error "Num type is used as Unit type")
      | value_unit Bool _ = raise (Error "Bool type is used as Unit type")
      | value_unit Location _ = raise (Error "Location type is used as Unit type")
      | value_unit Bot = raise (Error "not initialized")

    (* memory *)
    type memory = value Mem.t

    val emptyMemory = Mem.empty

    (* environment *)
    type env = (id, env_entry) Env.t
    and  env_entry = Addr of Loc.t | Proc of id * exp * env

    val emptyEnv = Env.empty

    fun env_loc e x =
        (case Env.lookup e x of
           Addr l => l
         | Proc _ => raise (Error "not allowed")) handle
          Env.Not_bound => raise (Error "not bound")

    fun env_proc e f = 
        (case Env.lookup e f of 
           Addr _ => raise (Error "not allowed") 
         | Proc p => p) handle 
          Env.Not_bound => raise (Error "not bound")

    fun opadd Num(a) Num(b) = Num(a+b)
	  | opadd Location(l) Num(o) = Location(Loc.addi l o)
	  | opadd Num(o) Location(l) = Location(Loc.addi l o)
	  | opadd _ _ = raise Error ("type Error")
	fun opsub Num(a) Num(b) = Num(a-b)
	  | opsub Location(l) Num(o) = Location(Loc.subi l o) 
	  | opsub Location(l1) Location(l2) = Num(Loc.sub l1 l2)
	  | opsub _ _ = raise Error ("type Error")
	fun opmul Num(i1) Num(i2) = Num(i1*i2)
	  | opmul _ _ = raise Error("type error")
	fun opdiv Num(i1) Num(0) = raise (Error "division by zero")
	  | opdiv Num(i1) Num(i2) = Num(i1/i2)
	  | opdiv _ _ = raise Error("type error")
	fun opequ Num(i1) Num(i2) = Bool(i1 = i2)
	  | opequ Bool(b1) Bool(b2) = Bool(b1 = b2)
	  | opequ Unit Unit = Bool(true)
	  | opequ Location(l1) Location(l2) = Bool(Loc.equal l1 l2) 
	  | opequ _ _ = raise Error("type error")
	fun oples Num(i1) Num(i2) = Bool(i1 < i2)
	  | oples Location(l1) Location(l2) = Bool(Loc.less l1 l2)  
	  | oples Unit Unit = Bool(false)
	  | oples _ _ = raise Error("type error")
	fun opnot Bool(b) = Bool(not b) 
	  | opnot _ = raise Error("type error")

	fun eval mem env exp = case exp of
		NUM(i) => (mem,Num(i))
		| TRUE => (mem,Bool(true))
		| FALSE => (mem,Bool(false))
		| UNIT => (mem,Unit)
		| VAR(id) =>
			let val loc = env_loc env id
				val value = Mem.load mem loc
			in (mem,value) end
		| ADD(e1,e2) => evalop mem env opadd e1 e2
		| SUB(e1,e2) => evalop mem env opsub e1 e2
		| MUL(e1,e2) => evalop mem env opmul e1 e2
 		| DIV(e1,e2) => evalop mem env opdiv e1 e2
		| EQUAL(e1,e2) => evalop mem env opequ e1 e2 
		| LESS(e1,e2) => evalop mem env oples e1 e2 
		| NOT(e) =>
			let val (mem',v) = eval mem env e
			in (mem', opnot v) end
		| ASSIGN(id,e) =>
			let val (mem',v) = eval mem env e
				val loc = env_loc env id
				val mem'' = Mem.store mem' loc v
			in (mem'',v) end
		| SEQ(e1,e2) =>
			let val (mem',_) = eval mem env e1
			in eval mem' env e2
			end
		| IF(e1,e2,e3) =>
			let val (mem',cond) = eval mem env e1
			in case cond of
				Bool(b) => if b then eval mem' env e2 else eval mem' env e3
				| _ => raise Error("type error")
			end
		| WHILE(e1,e2) =>
			let val (mem',cond) = eval mem env e1
			in case cond of 
				Bool(b) => if b then
						let val (mem'',_) = eval mem' env e2
						in eval mem'' env exp end
					else (mem',Unit) 
				| _ => raise Error("type error")
			end
		| FOR(id,e1,e2,e3) =>
			let val loc = env_loc env id
				val (mem',v1) = eval mem env e1
				val (mem'',v2) = eval mem' env e2
				val n1 = value_int v1
				val n2 = value_int v2
			in evalfor loc n1 n2 mem'' env e3 end
		| LETV(id,e1,e2) =>
			let val (mem',v) = eval mem env e1
				val (loc,mem'') = Mem.new mem'
				val env' = Env.bind env id Addr(loc)
				val mem2 = Mem.store mem'' loc v
			in eval mem2 env' e2 end
		| LETF(id,param,body,e) =>
			let val env' = Env.bind env id Proc(param,body,env)
			in eval mem env' e end
		| CALLV(id,e) =>
			let val (mem',v) = eval mem env e
				val (param,body,fenv) = env_proc env id
				val (loc,mem'') = Mem.new mem'
				val mem2 = Mem.store mem'' loc v
			in eval mem2 (Env.bind (Env.bind fenv id Proc(param,body,fenv)) param Addr(loc)) body end
		| CALLR(fid,aid) =>
			let val loc = env_loc env aid
				val (param,body,fenv) = env_proc env fid
			in eval mem (Env.bind (Env.bind fenv fid Proc(param,body,fenv)) param Addr(loc)) body
			end
		| READ(id) =>
			let val loc = env_loc env id
				val value = Num(read_int())
				val mem' = Mem.store mem loc value
			in (mem',value) end
		| WRITE(exp) =>
			let val (mem',v) = eval mem env exp
				val _ = print_int (value_int v)
				val _ = print_string "\n"
			in (mem',v) end
		| _ => raise Error("not implemented")  (* XXX *)

	(* for-loop evaluation procedure *)
	and evalfor loc n1 n2 mem env e =
		if n1 > n2 then (mem, Unit)
		else 
		let val mem' = Mem.store mem loc Num(n1)
			val (mem'',v) = eval mem' env e
		in evalfor loc (n1+1) (n2) mem'' env e end
	(* operation evaluation procedure *)
	and evalop mem env op e1 e2 =
		let val (mem',v1) = eval mem env e1
			val (mem'',v2) = eval mem' env e2
		in (mem'', op v1 v2) end

	fun run (mem, env, pgm) = let val (m,v) = eval mem env pgm in v end
  end
